/**
 * Copyright (c) 2019-2023 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
import { ParamDefinition as PD } from '../../../mol-util/param-definition';
import { ValueCell } from '../../../mol-util';
import { GeometryUtils } from '../geometry';
import { Sphere3D } from '../../../mol-math/geometry';
import { TextureImage } from '../../../mol-gl/renderable/util';
import { GroupMapping } from '../../util';
type TextAttachment = ('bottom-left' | 'bottom-center' | 'bottom-right' | 'middle-left' | 'middle-center' | 'middle-right' | 'top-left' | 'top-center' | 'top-right');
/** Text */
export interface Text {
    readonly kind: 'text';
    /** Number of characters in the text */
    charCount: number;
    /** Font Atlas */
    readonly fontTexture: ValueCell<TextureImage<Uint8Array>>;
    /** Center buffer as array of xyz values wrapped in a value cell */
    readonly centerBuffer: ValueCell<Float32Array>;
    /** Mapping buffer as array of xy values wrapped in a value cell */
    readonly mappingBuffer: ValueCell<Float32Array>;
    /** Depth buffer as array of z values wrapped in a value cell */
    readonly depthBuffer: ValueCell<Float32Array>;
    /** Index buffer as array of center index triplets wrapped in a value cell */
    readonly indexBuffer: ValueCell<Uint32Array>;
    /** Group buffer as array of group ids for each vertex wrapped in a value cell */
    readonly groupBuffer: ValueCell<Float32Array>;
    /** Texture coordinates buffer as array of uv values wrapped in a value cell */
    readonly tcoordBuffer: ValueCell<Float32Array>;
    /** Bounding sphere of the text */
    readonly boundingSphere: Sphere3D;
    /** Maps group ids to text indices */
    readonly groupMapping: GroupMapping;
    setBoundingSphere(boundingSphere: Sphere3D): void;
}
export declare namespace Text {
    function create(fontTexture: TextureImage<Uint8Array>, centers: Float32Array, mappings: Float32Array, depths: Float32Array, indices: Uint32Array, groups: Float32Array, tcoords: Float32Array, charCount: number, text?: Text): Text;
    function createEmpty(text?: Text): Text;
    const Params: {
        sizeFactor: PD.Numeric;
        borderWidth: PD.Numeric;
        borderColor: PD.Color;
        offsetX: PD.Numeric;
        offsetY: PD.Numeric;
        offsetZ: PD.Numeric;
        background: PD.BooleanParam;
        backgroundMargin: PD.Numeric;
        backgroundColor: PD.Color;
        backgroundOpacity: PD.Numeric;
        tether: PD.BooleanParam;
        tetherLength: PD.Numeric;
        tetherBaseWidth: PD.Numeric;
        attachment: PD.Select<TextAttachment>;
        fontFamily: PD.Select<import("./font-atlas").FontFamily>;
        fontQuality: PD.Select<number>;
        fontStyle: PD.Select<import("./font-atlas").FontStyle>;
        fontVariant: PD.Select<import("./font-atlas").FontVariant>;
        fontWeight: PD.Select<import("./font-atlas").FontWeight>;
        alpha: PD.Numeric;
        quality: PD.Select<"auto" | "medium" | "high" | "low" | "custom" | "highest" | "higher" | "lower" | "lowest">;
        material: PD.Group<PD.Normalize<{
            metalness: number;
            roughness: number;
            bumpiness: number;
        }>>;
        clip: PD.Group<PD.Normalize<{
            variant: import("../../../mol-util/clip").Clip.Variant;
            objects: PD.Normalize<{
                type: any;
                invert: any;
                position: any;
                rotation: any;
                scale: any;
            }>[];
        }>>;
        instanceGranularity: PD.BooleanParam;
    };
    type Params = typeof Params;
    const Utils: GeometryUtils<Text, Params>;
}
export {};
